from django.contrib import admin
from django_loci.admin import ObjectLocationInline

from openwisp_utils.admin import TimeReadonlyAdminMixin

from .models import Device


class DeviceAdmin(TimeReadonlyAdminMixin, admin.ModelAdmin):
    list_display = ('name', 'created', 'modified')
    save_on_top = True
    inlines = [ObjectLocationInline]


admin.site.register(Device, DeviceAdmin)



import os

from django.conf import settings
from django.core.files.storage import FileSystemStorage


class OverwriteStorage(FileSystemStorage):
    floorplan_upload_dir = 'floorplans'

    @classmethod
    def upload_to(cls, instance, filename):
        """
        passed to FloorPlan.image.upload_to
        """
        ext = filename.split('.')[-1]
        dir_ = cls.floorplan_upload_dir
        return '{0}/{1}.{2}'.format(dir_, instance.id, ext)

    def get_available_name(self, name, max_length=None):
        """
        removes file if it already exists
        """
        if self.exists(name):
            os.remove(os.path.join(settings.MEDIA_ROOT, name))
        return name



from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class LociConfig(AppConfig):
    name = 'django_loci'
    verbose_name = _('django-loci')

    def ready(self):
        import leaflet
        leaflet.app_settings['NO_GLOBALS'] = False
        self.load_receivers()

    def load_receivers(self):
        from .channels import receivers  # noqa



from django.contrib.auth import get_user_model
from django.test import TestCase

from ..models import FloorPlan, Location, ObjectLocation
from .base.test_admin_inline import BaseTestAdminInline
from .testdeviceapp.models import Device


class TestAdminInline(BaseTestAdminInline, TestCase):
    object_model = Device
    location_model = Location
    floorplan_model = FloorPlan
    object_location_model = ObjectLocation
    user_model = get_user_model()



from channels import Group
from channels.generic.websockets import WebsocketConsumer
from django.core.exceptions import ValidationError


location_broadcast_path = r'^/loci/location/(?P<pk>[^/]+)/$'


def _get_object_or_none(model, **kwargs):
    try:
        return model.objects.get(**kwargs)
    except (ValidationError, model.DoesNotExist):
        return None


class BaseLocationBroadcast(WebsocketConsumer):
    """
    Notifies that the coordinates of a location have changed
    to authorized users (superusers or organization operators)
    """
    http_user = True

    def connect(self, message, pk):
        location = _get_object_or_none(self.model, pk=pk)
        if not location or not self.is_authorized(message.user, location):
            message.reply_channel.send({'close': True})
            return
        message.reply_channel.send({'accept': True})
        Group('loci.mobile-location.{0}'.format(pk)).add(message.reply_channel)

    def is_authorized(self, user, location):
        return user.is_authenticated() and user.is_staff

    def disconnect(self, message, pk):
        """
        Perform things on connection close
        """
        Group('loci.mobile-location.{0}'.format(pk)).discard(message.reply_channel)



from .base.models import AbstractFloorPlan, AbstractLocation, AbstractObjectLocation


class Location(AbstractLocation):
    class Meta(AbstractLocation.Meta):
        abstract = False


class FloorPlan(AbstractFloorPlan):
    class Meta(AbstractFloorPlan.Meta):
        abstract = False


class ObjectLocation(AbstractObjectLocation):
    class Meta(AbstractObjectLocation.Meta):
        abstract = False



from django import forms
from django.templatetags.static import static
from leaflet.forms.widgets import LeafletWidget as BaseLeafletWidget

_floorplan_css = {'all': (static('django-loci/css/floorplan-widget.css'),)}


class ImageWidget(forms.FileInput):
    """
    Image widget which can show a thumbnail
    and carries information regarding
    the image width and height
    """
    template_name = 'admin/widgets/image.html'

    def __init__(self, *args, **kwargs):
        self.thumbnail = kwargs.pop('thumbnail', True)
        super(ImageWidget, self).__init__(*args, **kwargs)

    def get_context(self, name, value, attrs):
        c = super(ImageWidget, self).get_context(name, value, attrs)
        if value and hasattr(value, 'url'):
            c.update({
                'filename': value.name,
                'url': value.url,
                'thumbnail': self.thumbnail,
                'width': value.width,
                'height': value.height,
            })
        return c

    @property
    def media(self):
        css = _floorplan_css
        return forms.Media(css=css)


class FloorPlanWidget(forms.TextInput):
    """
    widget that allows to manage indoor coordinates
    """
    template_name = 'admin/widgets/floorplan.html'

    @property
    def media(self):
        js = (static('django-loci/js/floorplan-widget.js'),)
        css = _floorplan_css
        return forms.Media(js=js, css=css)


class LeafletWidget(BaseLeafletWidget):
    include_media = True
    geom_type = 'GEOMETRY'
    template_name = 'leaflet/admin/widget.html'
    modifiable = True
    map_width = '100%'
    map_height = '400px'
    display_raw = False
    settings_overrides = {}



from channels import Group
from django.db.models.signals import post_save
from django.dispatch import receiver

from ..models import Location


@receiver(post_save, sender=Location, dispatch_uid='ws_update_mobile_location')
def update_mobile_location(sender, instance, **kwargs):
    if not kwargs.get('created') and instance.geometry:
        group_name = 'loci.mobile-location.{0}'.format(str(instance.pk))
        message = {'text': instance.geometry.geojson}
        Group(group_name).send(message, immediately=True)



import json

from django import forms
from django.conf.urls import url
from django.contrib import admin
from django.contrib.contenttypes.admin import GenericStackedInline
from django.core.exceptions import ValidationError
from django.http import JsonResponse
from django.shortcuts import get_object_or_404
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _
from leaflet.admin import LeafletGeoAdmin

from openwisp_utils.admin import TimeReadonlyAdminMixin

from ..fields import GeometryField
from ..widgets import FloorPlanWidget, ImageWidget
from .models import AbstractLocation


class AbstractFloorPlanForm(forms.ModelForm):
    class Meta:
        exclude = tuple()
        widgets = {'image': ImageWidget()}

    class Media:
        css = {'all': ('django-loci/css/loci.css',)}


class AbstractFloorPlanAdmin(TimeReadonlyAdminMixin, admin.ModelAdmin):
    list_display = ['__str__', 'location', 'floor', 'created', 'modified']
    list_select_related = ['location']
    search_fields = ('location__name',)
    raw_id_fields = ('location',)
    save_on_top = True


class AbstractLocationForm(forms.ModelForm):
    class Meta:
        exclude = tuple()

    class Media:
        js = ('django-loci/js/loci.js',
              'django-loci/js/floorplan-inlines.js',)
        css = {'all': ('django-loci/css/loci.css',)}


class AbstractFloorPlanInline(TimeReadonlyAdminMixin, admin.StackedInline):
    extra = 0
    ordering = ('floor',)


class AbstractLocationAdmin(TimeReadonlyAdminMixin, LeafletGeoAdmin):
    list_display = ('name', 'short_type', 'is_mobile', 'created', 'modified')
    search_fields = ('name', 'address')
    list_filter = ('type', 'is_mobile')
    save_on_top = True

    def get_urls(self):
        app_label = self.model._meta.app_label
        return [
            url(r'^(?P<pk>[^/]+)/json/$',
                self.admin_site.admin_view(self.json_view),
                name='{0}_location_json'.format(app_label)),
            url(r'^(?P<pk>[^/]+)/floorplans/json/$',
                self.admin_site.admin_view(self.floorplans_json_view),
                name='{0}_location_floorplans_json'.format(app_label))
        ] + super(AbstractLocationAdmin, self).get_urls()

    def json_view(self, request, pk):
        instance = get_object_or_404(self.model, pk=pk)
        return JsonResponse({
            'name': instance.name,
            'type': instance.type,
            'is_mobile': instance.is_mobile,
            'address': instance.address,
            'geometry': json.loads(instance.geometry.json)
        })

    def floorplans_json_view(self, request, pk):
        instance = get_object_or_404(self.model, pk=pk)
        choices = []
        for floorplan in instance.floorplan_set.all():
            choices.append({
                'id': floorplan.pk,
                'str': str(floorplan),
                'floor': floorplan.floor,
                'image': floorplan.image.url,
                'image_width': floorplan.image.width,
                'image_height': floorplan.image.height,
            })
        return JsonResponse({'choices': choices})


class UnvalidatedChoiceField(forms.ChoiceField):
    """
    skips ChoiceField validation to allow custom options
    """
    def validate(self, value):
        super(forms.ChoiceField, self).validate(value)


_get_field = AbstractLocation._meta.get_field


class AbstractObjectLocationForm(forms.ModelForm):
    FORM_CHOICES = (
        ('', _('--- Please select an option ---')),
        ('new', _('New')),
        ('existing', _('Existing'))
    )
    LOCATION_TYPES = (
        FORM_CHOICES[0],
        AbstractLocation.LOCATION_TYPES[0],
        AbstractLocation.LOCATION_TYPES[1]
    )
    location_selection = forms.ChoiceField(choices=FORM_CHOICES, required=False)
    name = forms.CharField(label=_('Location name'),
                           max_length=75, required=False,
                           help_text=_get_field('name').help_text)
    address = forms.CharField(max_length=128, required=False)
    type = forms.ChoiceField(choices=LOCATION_TYPES, required=True,
                             help_text=_get_field('type').help_text)
    is_mobile = forms.BooleanField(label=_get_field('is_mobile').verbose_name,
                                   help_text=_get_field('is_mobile').help_text,
                                   required=False)
    geometry = GeometryField(required=False)
    floorplan_selection = forms.ChoiceField(required=False,
                                            choices=FORM_CHOICES)
    floorplan = UnvalidatedChoiceField(choices=((None, FORM_CHOICES[0][1]),),
                                       required=False)
    floor = forms.IntegerField(required=False)
    image = forms.ImageField(required=False,
                             widget=ImageWidget(thumbnail=False),
                             help_text=_('floor plan image'))
    indoor = forms.CharField(max_length=64, required=False,
                             label=_('indoor position'),
                             widget=FloorPlanWidget)

    class Meta:
        exclude = tuple()

    class Media:
        js = ('django-loci/js/loci.js',)
        css = {'all': ('django-loci/css/loci.css',)}

    def __init__(self, *args, **kwargs):
        super(AbstractObjectLocationForm, self).__init__(*args, **kwargs)
        # set initial values for custom fields
        initial = {}
        obj = self.instance
        location = obj.location
        floorplan = obj.floorplan
        if location:
            initial.update({
                'location_selection': 'existing',
                'type': location.type,
                'is_mobile': location.is_mobile,
                'name': location.name,
                'address': location.address,
                'geometry': location.geometry,
            })
        if floorplan:
            initial.update({
                'floorplan_selection': 'existing',
                'floorplan': floorplan.pk,
                'floor': floorplan.floor,
                'image': floorplan.image
            })
            floorplan_choices = self.fields['floorplan'].choices
            self.fields['floorplan'].choices = floorplan_choices + [(floorplan.pk, floorplan)]
        self.initial.update(initial)

    @cached_property
    def floorplan_model(self):
        return self.Meta.model.floorplan.field.rel.model

    @cached_property
    def location_model(self):
        return self.Meta.model.location.field.rel.model

    def clean_floorplan(self):
        floorplan_model = self.floorplan_model
        if self.cleaned_data['type'] != 'indoor' or self.cleaned_data['floorplan_selection'] == 'new':
            return None
        pk = self.cleaned_data['floorplan']
        if not pk:
            raise ValidationError(_('No floorplan selected'))
        try:
            fl = floorplan_model.objects.get(pk=pk)
        except floorplan_model.DoesNotExist:
            raise ValidationError(_('Selected floorplan does not exist'))
        if fl.location != self.cleaned_data['location']:
            raise ValidationError(_('This floorplan is associated to a different location'))
        return fl

    def clean(self):
        data = self.cleaned_data
        type_ = data['type']
        is_mobile = data['is_mobile']
        msg = _('this field is required for locations of type %(type)s')
        fields = []
        if not is_mobile and type_ in ['outdoor', 'indoor']:
            fields += ['location_selection', 'name', 'address', 'geometry']
        if not is_mobile and type_ == 'indoor':
            fields += ['floorplan_selection', 'floor', 'indoor']
            if data.get('floorplan_selection') == 'existing':
                fields.append('floorplan')
            elif data.get('floorplan_selection') == 'new':
                fields.append('image')
        elif is_mobile and not data.get('location'):
            data['name'] = ''
            data['address'] = ''
            data['geometry'] = ''
            data['location_selection'] = 'new'
        for field in fields:
            if field in data and data[field] in [None, '']:
                params = {'type': type_}
                err = ValidationError(msg, params=params)
                self.add_error(field, err)

    def _get_location_instance(self):
        data = self.cleaned_data
        location = data.get('location') or self.location_model()
        location.type = data.get('type') or location.type
        location.is_mobile = data.get('is_mobile') or location.is_mobile
        location.name = data.get('name') or location.name
        location.address = data.get('address') or location.address
        location.geometry = data.get('geometry') or location.geometry
        return location

    def _get_floorplan_instance(self):
        data = self.cleaned_data
        instance = self.instance
        floorplan = data.get('floorplan') or self.floorplan_model()
        floorplan.location = instance.location
        floor = data.get('floor')
        floorplan.floor = floor if floor is not None else floorplan.floor
        # the image path is updated only during creation
        # or if the image has been actually changed
        if data.get('image') and self.initial.get('image') != data.get('image'):
            floorplan.image = data['image']
        return floorplan

    def save(self, commit=True):
        instance = self.instance
        data = self.cleaned_data
        # create or update location
        instance.location = self._get_location_instance()
        # set name of mobile locations automatically
        if data['is_mobile'] and not instance.location.name:
            instance.location.name = str(self.instance.content_object)
        instance.location.save()
        # create or update floorplan
        if data['type'] == 'indoor':
            instance.floorplan = self._get_floorplan_instance()
            instance.floorplan.save()
        # call super
        return super(AbstractObjectLocationForm, self).save(commit=True)


class AbstractObjectLocationInline(TimeReadonlyAdminMixin, GenericStackedInline):
    verbose_name = _('geographic information')
    verbose_name_plural = verbose_name
    raw_id_fields = ('location',)
    max_num = 1
    extra = 1
    template = 'admin/django_loci/location_inline.html'
    fieldsets = (
        (None, {'fields': ('location_selection',)}),
        ('Geographic coordinates', {
            'classes': ('loci', 'coords'),
            'fields': ('location', 'type', 'is_mobile',
                       'name', 'address', 'geometry'),
        }),
        ('Indoor coordinates', {
            'classes': ('indoor', 'coords'),
            'fields': ('floorplan_selection', 'floorplan',
                       'floor', 'image', 'indoor',),
        })
    )



import logging
import os

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.contrib.gis.db import models
from django.contrib.humanize.templatetags.humanize import ordinal
from django.core.exceptions import ValidationError
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _

from openwisp_utils.base import TimeStampedEditableModel

from .. import settings as app_settings

logger = logging.getLogger(__name__)


@python_2_unicode_compatible
class AbstractLocation(TimeStampedEditableModel):
    LOCATION_TYPES = (
        ('outdoor', _('Outdoor environment (eg: street, square, garden, land)')),
        ('indoor', _('Indoor environment (eg: building, roofs, subway, large vehicles)')),
    )
    name = models.CharField(_('name'), max_length=75,
                            help_text=_('A descriptive name of the location '
                                        '(building name, company name, etc.)'))
    type = models.CharField(choices=LOCATION_TYPES, max_length=8, db_index=True,
                            help_text=_('indoor locations can have floorplans associated to them'))
    is_mobile = models.BooleanField(_('is mobile?'), default=False, db_index=True,
                                    help_text=_('is this location a moving object?'))
    address = models.CharField(_('address'), db_index=True,
                               max_length=256, blank=True)
    geometry = models.GeometryField(_('geometry'), blank=True, null=True)

    class Meta:
        abstract = True

    def __str__(self):
        return self.name

    def clean(self):
        self._validate_outdoor_floorplans()

    def _validate_outdoor_floorplans(self):
        """
        if a location type is changed from indoor to outdoor
        but the location has still floorplan associated to it,
        a ValidationError will be raised
        """
        if self.type == 'indoor' or self._state.adding:
            return
        if self.floorplan_set.count() > 0:
            msg = 'this location has floorplans associated to it, ' \
                  'please delete them before changing its type'
            raise ValidationError({'type': msg})

    @property
    def short_type(self):
        return _(self.type.capitalize())


@python_2_unicode_compatible
class AbstractFloorPlan(TimeStampedEditableModel):
    location = models.ForeignKey('django_loci.Location')
    floor = models.SmallIntegerField(_('floor'))
    image = models.ImageField(_('image'),
                              upload_to=app_settings.FLOORPLAN_STORAGE.upload_to,
                              storage=app_settings.FLOORPLAN_STORAGE(),
                              help_text=_('floor plan image'))

    class Meta:
        abstract = True
        unique_together = ('location', 'floor')

    def __str__(self):
        if self.floor is not 0:
            suffix = _('{0} floor').format(ordinal(self.floor))
        else:
            suffix = _('ground floor')
        return '{0} {1}'.format(self.location.name, suffix)

    def clean(self):
        self._validate_location_type()

    def delete(self, *args, **kwargs):
        super(AbstractFloorPlan, self).delete(*args, **kwargs)
        self._remove_image()

    def _validate_location_type(self):
        if self.location and self.location.type != 'indoor':
            msg = 'floorplans can only be associated '\
                  'to locations of type "indoor"'
            raise ValidationError(msg)

    def _remove_image(self):
        path = self.image.file.name
        if os.path.isfile(path):
            os.remove(path)
        else:
            msg = 'floorplan image not found while deleting {0}: {1}'
            logger.error(msg.format(self, path))


class AbstractObjectLocation(TimeStampedEditableModel):
    LOCATION_TYPES = (
        ('outdoor', _('Outdoor')),
        ('indoor', _('Indoor')),
        ('mobile', _('Mobile')),
    )
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.CharField(max_length=36, db_index=True)
    content_object = GenericForeignKey('content_type', 'object_id')
    location = models.ForeignKey('django_loci.Location', models.PROTECT,
                                 blank=True, null=True)
    floorplan = models.ForeignKey('django_loci.Floorplan', models.PROTECT,
                                  blank=True, null=True)
    indoor = models.CharField(_('indoor position'), max_length=64,
                              blank=True, null=True)

    class Meta:
        abstract = True
        unique_together = ('content_type', 'object_id')

    def _clean_indoor_location(self):
        """
        ensures related floorplan is not
        associated to a different location
        """
        # skip validation if the instance does not
        # have a floorplan assigned to it yet
        if not self.location or self.location.type != 'indoor' or not self.floorplan:
            return
        if self.location != self.floorplan.location:
            raise ValidationError(_('Invalid floorplan (belongs to a different location)'))

    def _raise_invalid_indoor(self):
        raise ValidationError({'indoor': _('invalid value')})

    def _clean_indoor_position(self):
        """
        ensures invalid indoor position values
        cannot be inserted into the database
        """
        # stop here if location not defined yet
        # (other validation errors will be triggered)
        if not self.location:
            return
        # do not allow non empty values for outdoor locations
        if self.location.type != 'indoor' and self.indoor not in [None, '']:
            self._raise_invalid_indoor()
        # allow empty values for outdoor locations
        elif self.location.type != 'indoor' and self.indoor in [None, '']:
            return
        # split indoor position
        position = []
        if self.indoor:
            position = self.indoor.split(',')
        # must have at least e elements
        if len(position) != 2:
            self._raise_invalid_indoor()
        # each member must be convertible to float
        else:
            for part in position:
                try:
                    float(part)
                except ValueError:
                    self._raise_invalid_indoor()

    def clean(self):
        self._clean_indoor_location()
        self._clean_indoor_position()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.5 on 2017-11-14 10:36
from __future__ import unicode_literals

from django.db import migrations, models
import django.utils.timezone
import model_utils.fields
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Device',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('name', models.CharField(max_length=75, verbose_name='name')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



import os

from django.core.exceptions import ValidationError

from .. import TestLociMixin


class BaseTestModels(TestLociMixin):
    def test_location_str(self):
        loc = self.location_model(name='test-location')
        self.assertEqual(str(loc), loc.name)

    def test_floorplan_str(self):
        loc = self._create_location()
        fl = self.floorplan_model(location=loc, floor=2)
        self.assertEqual(str(fl), 'test-location 2nd floor')
        fl.floor = 0
        self.assertEqual(str(fl), 'test-location ground floor')

    def test_object_location_clean_location(self):
        l1 = self._create_location(type='indoor')
        l2 = self._create_location(type='indoor')
        fl2 = self._create_floorplan(location=l2)
        obj = self._create_object()
        ol = self.object_location_model(content_object=obj,
                                        location=l1,
                                        floorplan=fl2)
        try:
            ol.full_clean()
        except ValidationError as e:
            self.assertIn('__all__', e.message_dict)
            self.assertEqual(e.message_dict.get('__all__')[0],
                             'Invalid floorplan (belongs to a different location)')
        else:
            self.fail('ValidationError not raised')

    def test_floorplan_image(self):
        fl = self._create_floorplan()
        path = fl.image.file.name.split('/')
        name = path[-1]
        dir_ = path[-2]
        self.assertEqual(name, '{0}.jpg'.format(fl.id))
        self.assertEqual(dir_, 'floorplans')
        # overwrite
        fl.image = self._get_simpleuploadedfile()
        fl.full_clean()
        fl.save()
        path = fl.image.file.name.split('/')
        name = path[-1]
        dir_ = path[-2]
        self.assertEqual(name, '{0}.jpg'.format(fl.id))
        self.assertEqual(dir_, 'floorplans')
        # delete
        fl.delete()
        self.assertFalse(os.path.isfile(fl.image.file.name))

    def test_floorplan_delete_corner_case(self):
        fl = self._create_floorplan()
        os.remove(fl.image.file.name)
        # there should be no failure
        fl.delete()

    def test_floorplan_association_validation(self):
        outdoor = self._create_location(type='outdoor')
        try:
            self._create_floorplan(location=outdoor)
        except ValidationError as e:
            err_str = str(e)
            self.assertIn('floorplans can only be associated to', err_str)
            self.assertIn('indoor', err_str)
        else:
            self.fail('ValidationError not raised')

    def test_location_change_indoor_to_outdoor(self):
        fl = self._create_floorplan()
        location = fl.location
        location.type = 'outdoor'
        try:
            location.full_clean()
        except ValidationError as e:
            self.assertIn('type', e.message_dict)
            err_str = str(e.message_dict['type'])
            self.assertIn('this location has floorplans', err_str)
            self.assertIn('please delete them', err_str)
        else:
            self.fail('ValidationError not raised')

    def _test_indoor_position_validation_error(self, ol):
        try:
            ol.full_clean()
        except ValidationError as e:
            self.assertIn('indoor', e.message_dict)
            self.assertIn('invalid value', e.message_dict['indoor'])
        else:
            self.fail('ValidationError not raised')

    def test_invalid_indoor_position(self):
        loc = self._create_location(type='indoor')
        ol = self._create_object_location(location=loc)
        ol.indoor = 'TOTALLYWRONG'
        self._test_indoor_position_validation_error(ol)
        ol.indoor = 'WRONG,WRONG'
        self._test_indoor_position_validation_error(ol)
        ol.indoor = '10,WRONG'
        self._test_indoor_position_validation_error(ol)
        ol.indoor = 'WRONG,10'
        self._test_indoor_position_validation_error(ol)
        ol.indoor = '10,10.10,100'
        self._test_indoor_position_validation_error(ol)
        ol.indoor = 'TOTALLY.WRONG'
        self._test_indoor_position_validation_error(ol)
        ol.indoor = ''
        self._test_indoor_position_validation_error(ol)
        ol.indoor = None
        self._test_indoor_position_validation_error(ol)
        ol.indoor = '100.2300,-45.23454'
        ol.full_clean()
        # outdoor allows empty but not invalid values
        loc.type = 'outdoor'
        loc.full_clean()
        loc.save()
        ol.indoor = None
        ol.full_clean()
        ol.indoor = ''
        ol.full_clean()
        ol.indoor = 'TOTALLY.WRONG'
        self._test_indoor_position_validation_error(ol)
        # outdoor does not allow valid indoor positions
        ol.indoor = '100.2300,-45.23454'
        self._test_indoor_position_validation_error(ol)



import os

from django.conf import settings
from django.contrib.gis.geos import GEOSGeometry
from django.db.models.fields.files import ImageFieldFile
from django.urls import reverse

from .. import TestAdminMixin, TestLociMixin


class BaseTestAdminInline(TestAdminMixin, TestLociMixin):
    app_label = 'django_loci'
    inline_field_prefix = 'objectlocation-content_type-object_id'
    _p = '{0}-{1}'.format(app_label, inline_field_prefix)
    _params = {
        '{0}-0-is_mobile'.format(_p): False,
        '{0}-0-name'.format(_p): 'Centro Piazza Venezia',
        '{0}-0-address'.format(_p): 'Piazza Venezia, Roma, Italia',
        '{0}-0-geometry'.format(_p): '{"type": "Point", "coordinates": [12.512124, 41.898903]}',
        '{0}-TOTAL_FORMS'.format(_p): '1',
        '{0}-INITIAL_FORMS'.format(_p): '0',
        '{0}-MIN_NUM_FORMS'.format(_p): '0',
        '{0}-MAX_NUM_FORMS'.format(_p): '1',
    }

    def test_json_urls(self):
        self._login_as_admin()
        r = self.client.get(reverse('admin:testdeviceapp_device_add'))
        url = reverse('admin:django_loci_location_json', args=['0000'])
        self.assertContains(r, url)
        url = reverse('admin:django_loci_location_floorplans_json', args=['0000'])
        self.assertContains(r, url)

    def test_add_outdoor_new(self):
        self._login_as_admin()
        p = self._p
        params = self._params.copy()
        params.update({
            'name': 'test-outdoor-add-new',
            '{0}-0-type'.format(p): 'outdoor',
            '{0}-0-location_selection'.format(p): 'new',
            '{0}-0-location'.format(p): '',
            '{0}-0-floorplan_selection'.format(p): '',
            '{0}-0-floorplan'.format(p): '',
            '{0}-0-floor'.format(p): '',
            '{0}-0-image'.format(p): '',
            '{0}-0-indoor'.format(p): '',
            '{0}-0-id'.format(p): '',
        })
        r = self.client.post(reverse('admin:testdeviceapp_device_add'), params, follow=True)
        self.assertNotContains(r, 'errors')
        loc = self.location_model.objects.get(name=params['{0}-0-name'.format(p)])
        self.assertEqual(loc.address, params['{0}-0-address'.format(p)])
        self.assertEqual(loc.geometry.coords, GEOSGeometry(params['{0}-0-geometry'.format(p)]).coords)
        self.assertEqual(loc.objectlocation_set.count(), 1)
        self.assertEqual(loc.objectlocation_set.first().content_object.name, params['name'])

    def test_add_outdoor_existing(self):
        self._login_as_admin()
        p = self._p
        pre_loc = self._create_location()
        params = self._params.copy()
        params.update({
            'name': 'test-outdoor-add-existing',
            '{0}-0-type'.format(p): 'outdoor',
            '{0}-0-location_selection'.format(p): 'existing',
            '{0}-0-location'.format(p): pre_loc.id,
            '{0}-0-floorplan_selection'.format(p): '',
            '{0}-0-floorplan'.format(p): '',
            '{0}-0-floor'.format(p): '',
            '{0}-0-image'.format(p): '',
            '{0}-0-indoor'.format(p): '',
            '{0}-0-id'.format(p): '',
        })
        r = self.client.post(reverse('admin:testdeviceapp_device_add'), params, follow=True)
        self.assertNotContains(r, 'errors')
        loc = self.location_model.objects.get(name=params['{0}-0-name'.format(p)])
        self.assertEqual(pre_loc.id, loc.id)
        self.assertEqual(loc.address, params['{0}-0-address'.format(p)])
        self.assertEqual(loc.geometry.coords, GEOSGeometry(params['{0}-0-geometry'.format(p)]).coords)
        self.assertEqual(loc.objectlocation_set.count(), 1)
        self.assertEqual(loc.objectlocation_set.first().content_object.name, params['name'])
        self.assertEqual(self.location_model.objects.count(), 1)

    def test_change_outdoor(self):
        self._login_as_admin()
        p = self._p
        obj = self._create_object(name='test-change-outdoor')
        pre_loc = self._create_location()
        ol = self._create_object_location(location=pre_loc,
                                          content_object=obj)
        # -- ensure change form doesn't raise any exception
        r = self.client.get(reverse('admin:testdeviceapp_device_change', args=[obj.pk]))
        self.assertContains(r, obj.name)
        # -- post changes
        params = self._params.copy()
        params.update({
            'name': obj.name,
            '{0}-0-type'.format(p): 'outdoor',
            '{0}-0-location_selection'.format(p): 'existing',
            '{0}-0-location'.format(p): pre_loc.id,
            '{0}-0-floorplan_selection'.format(p): '',
            '{0}-0-floorplan'.format(p): '',
            '{0}-0-floor'.format(p): '',
            '{0}-0-image'.format(p): '',
            '{0}-0-indoor'.format(p): '',
            '{0}-0-id'.format(p): ol.id,
            '{0}-INITIAL_FORMS'.format(p): '1',
        })
        r = self.client.post(reverse('admin:testdeviceapp_device_change', args=[obj.pk]), params, follow=True)
        self.assertNotContains(r, 'errors')
        loc = self.location_model.objects.get(name=params['{0}-0-name'.format(p)])
        self.assertEqual(pre_loc.id, loc.id)
        self.assertEqual(loc.address, params['{0}-0-address'.format(p)])
        self.assertEqual(loc.geometry.coords, GEOSGeometry(params['{0}-0-geometry'.format(p)]).coords)
        self.assertEqual(loc.objectlocation_set.count(), 1)
        self.assertEqual(loc.objectlocation_set.first().content_object.name, params['name'])
        self.assertEqual(self.location_model.objects.count(), 1)

    def test_change_outdoor_to_different_location(self):
        self._login_as_admin()
        p = self._p
        ol = self._create_object_location()
        new_loc = self._create_location(name='different-location',
                                        address='Piazza Venezia, Roma, Italia',
                                        geometry='SRID=4326;POINT (12.512324 41.898703)')
        # -- post changes
        params = self._params.copy()
        changed_name = '{0} changed'.format(new_loc.name)
        params.update({
            'name': 'test-outdoor-change-different',
            '{0}-0-type'.format(p): 'outdoor',
            '{0}-0-location_selection'.format(p): 'existing',
            '{0}-0-location'.format(p): new_loc.id,
            '{0}-0-name'.format(p): changed_name,
            '{0}-0-address'.format(p): new_loc.address,
            '{0}-0-geometry'.format(p): new_loc.geometry.geojson,
            '{0}-0-floorplan_selection'.format(p): '',
            '{0}-0-floorplan'.format(p): '',
            '{0}-0-floor'.format(p): '',
            '{0}-0-image'.format(p): '',
            '{0}-0-indoor'.format(p): '',
            '{0}-0-id'.format(p): ol.id,
            '{0}-INITIAL_FORMS'.format(p): '1',
        })
        r = self.client.post(reverse('admin:testdeviceapp_device_change',
                                     args=[ol.content_object.pk]),
                             params, follow=True)
        self.assertNotContains(r, 'errors')
        loc = self.location_model.objects.get(name=changed_name)
        self.assertEqual(new_loc.id, loc.id)
        self.assertEqual(loc.address, params['{0}-0-address'.format(p)])
        self.assertEqual(loc.geometry.coords, GEOSGeometry(params['{0}-0-geometry'.format(p)]).coords)
        self.assertEqual(loc.objectlocation_set.count(), 1)
        self.assertEqual(loc.objectlocation_set.first().content_object.name, params['name'])
        self.assertEqual(self.location_model.objects.count(), 2)

    def test_add_indoor_new_location_new_floorplan(self):
        self._login_as_admin()
        p = self._p
        params = self._params.copy()
        floorplan_file = open(os.path.join(settings.BASE_DIR, 'media/floorplan.jpg'), 'rb')
        params.update({
            'name': 'test-add-indoor-new-location-new-floorplan',
            '{0}-0-type'.format(p): 'indoor',
            '{0}-0-location_selection'.format(p): 'new',
            '{0}-0-location'.format(p): '',
            '{0}-0-floorplan_selection'.format(p): 'new',
            '{0}-0-floorplan'.format(p): '',
            '{0}-0-floor'.format(p): '1',
            '{0}-0-image'.format(p): floorplan_file,
            '{0}-0-indoor'.format(p): '-100,100',
            '{0}-0-id'.format(p): '',
        })
        r = self.client.post(reverse('admin:testdeviceapp_device_add'), params, follow=True)
        floorplan_file.close()
        self.assertNotContains(r, 'errors')
        loc = self.location_model.objects.get(name=params['{0}-0-name'.format(p)])
        self.assertEqual(loc.address, params['{0}-0-address'.format(p)])
        self.assertEqual(loc.geometry.coords, GEOSGeometry(params['{0}-0-geometry'.format(p)]).coords)
        self.assertEqual(loc.objectlocation_set.count(), 1)
        self.assertEqual(self.location_model.objects.count(), 1)
        self.assertEqual(self.floorplan_model.objects.count(), 1)
        ol = loc.objectlocation_set.first()
        self.assertEqual(ol.content_object.name, params['name'])
        self.assertEqual(ol.location.type, 'indoor')
        self.assertEqual(ol.floorplan.floor, 1)
        self.assertIsInstance(ol.floorplan.image, ImageFieldFile)
        self.assertEqual(ol.indoor, '-100,100')

    def test_add_indoor_existing_location_new_floorplan(self):
        self._login_as_admin()
        obj = self._create_object(name='test-add-indoor-existing-location-new-floorplan')
        pre_loc = self._create_location(type='indoor')
        p = self._p
        params = self._params.copy()
        floorplan_file = open(os.path.join(settings.BASE_DIR, 'media/floorplan.jpg'), 'rb')
        params.update({
            'name': obj.name,
            '{0}-0-type'.format(p): 'indoor',
            '{0}-0-location_selection'.format(p): 'existing',
            '{0}-0-location'.format(p): pre_loc.id,
            '{0}-0-name'.format(p): pre_loc.name,
            '{0}-0-address'.format(p): pre_loc.address,
            '{0}-0-geometry'.format(p): pre_loc.geometry.geojson,
            '{0}-0-floorplan_selection'.format(p): 'new',
            '{0}-0-floorplan'.format(p): '',
            '{0}-0-floor'.format(p): '0',
            '{0}-0-image'.format(p): floorplan_file,
            '{0}-0-indoor'.format(p): '-100,100',
            '{0}-0-id'.format(p): '',
        })
        r = self.client.post(reverse('admin:testdeviceapp_device_add'), params, follow=True)
        floorplan_file.close()
        self.assertNotContains(r, 'errors')
        loc = self.location_model.objects.get(name=params['{0}-0-name'.format(p)])
        self.assertEqual(loc.address, params['{0}-0-address'.format(p)])
        self.assertEqual(loc.geometry.coords, GEOSGeometry(params['{0}-0-geometry'.format(p)]).coords)
        self.assertEqual(loc.objectlocation_set.count(), 1)
        self.assertEqual(self.location_model.objects.count(), 1)
        self.assertEqual(self.floorplan_model.objects.count(), 1)
        ol = loc.objectlocation_set.first()
        self.assertEqual(ol.content_object.name, params['name'])
        self.assertEqual(ol.location.type, 'indoor')
        self.assertEqual(ol.floorplan.floor, 0)
        self.assertIsInstance(ol.floorplan.image, ImageFieldFile)
        self.assertEqual(ol.indoor, '-100,100')

    def test_add_indoor_existing_location_existing_floorplan(self):
        self._login_as_admin()
        obj = self._create_object(name='test-add-indoor-existing-location-new-floorplan')
        pre_loc = self._create_location(type='indoor')
        pre_fl = self._create_floorplan(location=pre_loc, floor=2)
        p = self._p
        params = self._params.copy()
        changed_name = '{0} changed'.format(pre_loc.name)
        params.update({
            'name': obj.name,
            '{0}-0-type'.format(p): 'indoor',
            '{0}-0-location_selection'.format(p): 'existing',
            '{0}-0-location'.format(p): pre_loc.id,
            '{0}-0-name'.format(p): changed_name,
            '{0}-0-address'.format(p): pre_loc.address,
            '{0}-0-location-geometry'.format(p): pre_loc.geometry,
            '{0}-0-floorplan_selection'.format(p): 'existing',
            '{0}-0-floorplan'.format(p): pre_fl.id,
            '{0}-0-floor'.format(p): 3,  # floor
            '{0}-0-image'.format(p): '',
            '{0}-0-indoor'.format(p): '-110,110',
            '{0}-0-id'.format(p): '',
        })
        r = self.client.post(reverse('admin:testdeviceapp_device_add'), params, follow=True)
        self.assertNotContains(r, 'errors')
        loc = self.location_model.objects.get(name=changed_name)
        self.assertEqual(loc.id, pre_loc.id)
        self.assertEqual(loc.address, params['{0}-0-address'.format(p)])
        self.assertEqual(loc.geometry.coords, GEOSGeometry(params['{0}-0-geometry'.format(p)]).coords)
        self.assertEqual(loc.objectlocation_set.count(), 1)
        self.assertEqual(self.location_model.objects.count(), 1)
        self.assertEqual(self.floorplan_model.objects.count(), 1)
        ol = loc.objectlocation_set.first()
        self.assertEqual(ol.content_object.name, params['name'])
        self.assertEqual(ol.location.type, 'indoor')
        self.assertEqual(ol.floorplan.id, pre_fl.id)
        self.assertEqual(ol.floorplan.floor, 3)
        self.assertIsInstance(ol.floorplan.image, ImageFieldFile)
        self.assertEqual(ol.indoor, '-110,110')

    def test_change_indoor(self):
        self._login_as_admin()
        p = self._p
        obj = self._create_object(name='test-change-indoor')
        pre_loc = self._create_location(type='indoor')
        pre_fl = self._create_floorplan(location=pre_loc)
        ol = self._create_object_location(content_object=obj,
                                          location=pre_loc,
                                          floorplan=pre_fl,
                                          indoor='-100,100')
        # -- ensure change form doesn't raise any exception
        r = self.client.get(reverse('admin:testdeviceapp_device_change', args=[obj.pk]))
        self.assertContains(r, obj.name)
        # -- post changes
        params = self._params.copy()
        floorplan_file = open(os.path.join(settings.BASE_DIR, 'media/floorplan.jpg'), 'rb')
        changed_name = '{0} changed'.format(pre_loc.name)
        params.update({
            'name': obj.name,
            '{0}-0-type'.format(p): 'indoor',
            '{0}-0-location_selection'.format(p): 'existing',
            '{0}-0-location'.format(p): pre_loc.id,
            '{0}-0-name'.format(p): changed_name,
            '{0}-0-address'.format(p): 'changed-address',
            '{0}-0-location-geometry'.format(p): pre_loc.geometry,
            '{0}-0-floorplan_selection'.format(p): 'existing',
            '{0}-0-floorplan'.format(p): pre_fl.id,
            '{0}-0-floor'.format(p): 3,  # floor
            '{0}-0-image'.format(p): floorplan_file,
            '{0}-0-indoor'.format(p): '-110,110',
            '{0}-0-id'.format(p): ol.id,
            '{0}-INITIAL_FORMS'.format(p): '1',
        })
        r = self.client.post(reverse('admin:testdeviceapp_device_change',
                                     args=[obj.pk]),
                             params, follow=True)
        floorplan_file.close()
        self.assertNotContains(r, 'errors')
        loc = self.location_model.objects.get(name=changed_name)
        self.assertEqual(loc.id, pre_loc.id)
        self.assertEqual(loc.address, 'changed-address')
        self.assertEqual(loc.geometry.coords, GEOSGeometry(params['{0}-0-geometry'.format(p)]).coords)
        self.assertEqual(loc.objectlocation_set.count(), 1)
        self.assertEqual(self.location_model.objects.count(), 1)
        self.assertEqual(self.floorplan_model.objects.count(), 1)
        ol = loc.objectlocation_set.first()
        self.assertEqual(ol.content_object.name, params['name'])
        self.assertEqual(ol.location.type, 'indoor')
        self.assertEqual(ol.floorplan.id, pre_fl.id)
        self.assertEqual(ol.floorplan.floor, 3)
        self.assertIsInstance(ol.floorplan.image, ImageFieldFile)
        self.assertEqual(ol.indoor, '-110,110')

    def test_change_indoor_missing_indoor_position(self):
        self._login_as_admin()
        p = self._p
        obj = self._create_object(name='test-change-indoor')
        pre_loc = self._create_location(type='indoor')
        pre_fl = self._create_floorplan(location=pre_loc)
        ol = self._create_object_location(content_object=obj,
                                          location=pre_loc,
                                          floorplan=pre_fl,
                                          indoor='-100,100')
        # -- post changes
        params = self._params.copy()
        params.update({
            'name': obj.name,
            '{0}-0-type'.format(p): 'indoor',
            '{0}-0-location_selection'.format(p): 'existing',
            '{0}-0-location'.format(p): pre_loc.id,
            '{0}-0-name'.format(p): pre_loc.name,
            '{0}-0-address'.format(p): pre_loc.address,
            '{0}-0-location-geometry'.format(p): pre_loc.geometry,
            '{0}-0-floorplan_selection'.format(p): 'existing',
            '{0}-0-floorplan'.format(p): pre_fl.id,
            '{0}-0-floor'.format(p): pre_fl.floor,
            '{0}-0-indoor'.format(p): '',
            '{0}-0-id'.format(p): ol.id,
            '{0}-INITIAL_FORMS'.format(p): '1',
        })
        r = self.client.post(reverse('admin:testdeviceapp_device_change',
                                     args=[obj.pk]),
                             params, follow=True)
        self.assertContains(r, 'errors field-indoor')

    def test_add_outdoor_invalid(self):
        self._login_as_admin()
        p = self._p
        params = self._params.copy()
        params.update({
            'name': 'test-outdoor-invalid',
            '{0}-0-type'.format(p): 'outdoor',
            '{0}-0-location_selection'.format(p): 'new',
            '{0}-0-location'.format(p): '',
            '{0}-0-name'.format(p): '',
            '{0}-0-address'.format(p): '',
            '{0}-0-geometry'.format(p): '',
        })
        r = self.client.post(reverse('admin:testdeviceapp_device_add'), params, follow=True)
        self.assertContains(r, 'errors field-name')
        self.assertContains(r, 'errors field-address')
        self.assertContains(r, 'errors field-geometry')

    def test_add_outdoor_invalid_geometry(self):
        self._login_as_admin()
        p = self._p
        params = self._params.copy()
        params.update({
            'name': 'test-outdoor-invalid-geometry',
            '{0}-0-type'.format(p): 'outdoor',
            '{0}-0-location_selection'.format(p): 'new',
            '{0}-0-location'.format(p): '',
            '{0}-0-geometry'.format(p): 'INVALID',
        })
        r = self.client.post(reverse('admin:testdeviceapp_device_add'), params, follow=True)
        self.assertContains(r, 'errors field-geometry')

    def test_add_mobile(self):
        self._login_as_admin()
        p = self._p
        params = self._params.copy()
        params.update({
            'name': 'test-add-mobile',
            '{0}-0-type'.format(p): 'outdoor',
            '{0}-0-is_mobile'.format(p): True,
            '{0}-0-location_selection'.format(p): 'new',
            '{0}-0-name'.format(p): '',
            '{0}-0-address'.format(p): '',
            '{0}-0-geometry'.format(p): '',
        })
        self.assertEqual(self.location_model.objects.count(), 0)
        r = self.client.post(reverse('admin:testdeviceapp_device_add'), params, follow=True)
        self.assertNotContains(r, 'errors')
        self.assertEqual(self.location_model.objects.filter(is_mobile=True).count(), 1)
        self.assertEqual(self.object_location_model.objects.count(), 1)
        loc = self.location_model.objects.first()
        self.assertEqual(loc.objectlocation_set.first().content_object.name, params['name'])
        self.assertEqual(loc.name, params['name'])

    def test_change_mobile(self):
        self._login_as_admin()
        obj = self._create_object(name='test-change-mobile')
        pre_loc = self._create_location(name=obj.name, is_mobile=True)
        ol = self._create_object_location(content_object=obj,
                                          location=pre_loc)
        p = self._p
        params = self._params.copy()
        params.update({
            'name': 'test-change-mobile',
            '{0}-0-type'.format(p): 'outdoor',
            '{0}-0-is_mobile'.format(p): True,
            '{0}-0-location'.format(p): pre_loc.id,
            '{0}-0-name'.format(p): '',
            '{0}-0-address'.format(p): '',
            '{0}-0-geometry'.format(p): '',
            '{0}-0-location_selection'.format(p): 'new',
            '{0}-0-id'.format(p): ol.id,
            '{0}-INITIAL_FORMS'.format(p): '1',
        })
        self.assertEqual(self.location_model.objects.count(), 1)
        r = self.client.post(reverse('admin:testdeviceapp_device_change',
                                     args=[obj.pk]),
                             params, follow=True)
        self.assertNotContains(r, 'errors')
        self.assertEqual(self.object_location_model.objects.count(), 1)
        self.assertEqual(self.location_model.objects.filter(is_mobile=True).count(), 1)
        loc = self.location_model.objects.first()
        self.assertEqual(loc.objectlocation_set.first().content_object.name, params['name'])

    def test_change_indoor_missing_floorplan_pk(self):
        self._login_as_admin()
        p = self._p
        obj = self._create_object(name='test-floorplan-error')
        pre_loc = self._create_location(type='indoor')
        pre_fl = self._create_floorplan(location=pre_loc)
        ol = self._create_object_location(content_object=obj,
                                          location=pre_loc,
                                          floorplan=pre_fl,
                                          indoor='-100,100')
        # -- post changes
        params = self._params.copy()
        params.update({
            'name': obj.name,
            '{0}-0-type'.format(p): 'indoor',
            '{0}-0-location_selection'.format(p): 'existing',
            '{0}-0-location'.format(p): pre_loc.id,
            '{0}-0-name'.format(p): pre_loc.name,
            '{0}-0-type'.format(p): pre_loc.type,
            '{0}-0-is_mobile'.format(p): pre_loc.is_mobile,
            '{0}-0-address'.format(p): pre_loc.address,
            '{0}-0-location-geometry'.format(p): pre_loc.geometry,
            '{0}-0-floorplan_selection'.format(p): 'existing',
            '{0}-0-floorplan'.format(p): '',
            '{0}-0-floor'.format(p): pre_fl.floor,
            '{0}-0-indoor'.format(p): '-100,100',
            '{0}-0-id'.format(p): ol.id,
            '{0}-INITIAL_FORMS'.format(p): '1',
        })
        r = self.client.post(reverse('admin:testdeviceapp_device_change',
                                     args=[obj.pk]),
                             params, follow=True)
        self.assertContains(r, 'errors field-floorplan')
        self.assertContains(r, 'No floorplan selected')

    def test_change_indoor_floorplan_doesnotexist(self):
        self._login_as_admin()
        p = self._p
        obj = self._create_object(name='test-floorplan-error')
        pre_loc = self._create_location(type='indoor')
        pre_fl = self._create_floorplan(location=pre_loc)
        ol = self._create_object_location(content_object=obj,
                                          location=pre_loc,
                                          floorplan=pre_fl,
                                          indoor='-100,100')
        # -- post changes
        params = self._params.copy()
        params.update({
            'name': obj.name,
            '{0}-0-type'.format(p): 'indoor',
            '{0}-0-location_selection'.format(p): 'existing',
            '{0}-0-location'.format(p): pre_loc.id,
            '{0}-0-name'.format(p): pre_loc.name,
            '{0}-0-type'.format(p): pre_loc.type,
            '{0}-0-is_mobile'.format(p): pre_loc.is_mobile,
            '{0}-0-address'.format(p): pre_loc.address,
            '{0}-0-location-geometry'.format(p): pre_loc.geometry,
            '{0}-0-floorplan_selection'.format(p): 'existing',
            '{0}-0-floorplan'.format(p): self.floorplan_model().id,
            '{0}-0-floor'.format(p): pre_fl.floor,
            '{0}-0-indoor'.format(p): '-100,100',
            '{0}-0-id'.format(p): ol.id,
            '{0}-INITIAL_FORMS'.format(p): '1',
        })
        r = self.client.post(reverse('admin:testdeviceapp_device_change',
                                     args=[obj.pk]),
                             params, follow=True)
        self.assertContains(r, 'errors field-floorplan')
        self.assertContains(r, 'Selected floorplan does not exist')

    def test_change_indoor_floorplan_different_location(self):
        self._login_as_admin()
        p = self._p
        obj = self._create_object(name='test-floorplan-error')
        pre_loc = self._create_location(type='indoor')
        pre_fl = self._create_floorplan(location=pre_loc)
        ol = self._create_object_location(content_object=obj,
                                          location=pre_loc,
                                          floorplan=pre_fl,
                                          indoor='-100,100')
        fl = self._create_floorplan()
        # -- post changes
        params = self._params.copy()
        params.update({
            'name': obj.name,
            '{0}-0-type'.format(p): 'indoor',
            '{0}-0-location_selection'.format(p): 'existing',
            '{0}-0-location'.format(p): pre_loc.id,
            '{0}-0-name'.format(p): pre_loc.name,
            '{0}-0-address'.format(p): pre_loc.address,
            '{0}-0-location-geometry'.format(p): pre_loc.geometry,
            '{0}-0-floorplan_selection'.format(p): 'existing',
            '{0}-0-floorplan'.format(p): fl.id,
            '{0}-0-floor'.format(p): pre_fl.floor,
            '{0}-0-indoor'.format(p): '-100,100',
            '{0}-0-id'.format(p): ol.id,
            '{0}-INITIAL_FORMS'.format(p): '1',
        })
        r = self.client.post(reverse('admin:testdeviceapp_device_change',
                                     args=[obj.pk]),
                             params, follow=True)
        self.assertContains(r, 'errors field-floorplan')
        self.assertContains(r, 'This floorplan is associated to a different location')



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2017-11-25 10:09
from __future__ import unicode_literals

import django.contrib.gis.db.models.fields
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import django_loci.storage
import model_utils.fields
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='FloorPlan',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('floor', models.SmallIntegerField(verbose_name='floor')),
                ('image', models.ImageField(help_text='floor plan image', storage=django_loci.storage.OverwriteStorage(), upload_to=django_loci.storage.OverwriteStorage.upload_to, verbose_name='image')),
            ],
        ),
        migrations.CreateModel(
            name='Location',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('name', models.CharField(help_text='A descriptive name of the location (building name, company name, etc.)', max_length=75, verbose_name='name')),
                ('type', models.CharField(choices=[('outdoor', 'Outdoor environment (eg: street, square, garden, land)'), ('indoor', 'Indoor environment (eg: building, roofs, subway, large vehicles)')], db_index=True, help_text='indoor locations can have floorplans associated to them', max_length=8)),
                ('is_mobile', models.BooleanField(db_index=True, default=False, help_text='is this location a moving object?', verbose_name='is mobile?')),
                ('address', models.CharField(blank=True, db_index=True, max_length=256, verbose_name='address')),
                ('geometry', django.contrib.gis.db.models.fields.GeometryField(blank=True, null=True, srid=4326, verbose_name='geometry')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ObjectLocation',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('object_id', models.CharField(db_index=True, max_length=36)),
                ('indoor', models.CharField(blank=True, max_length=64, null=True, verbose_name='indoor position')),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
                ('floorplan', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='django_loci.FloorPlan')),
                ('location', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='django_loci.Location')),
            ],
        ),
        migrations.AddField(
            model_name='floorplan',
            name='location',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_loci.Location'),
        ),
        migrations.AlterUniqueTogether(
            name='objectlocation',
            unique_together=set([('content_type', 'object_id')]),
        ),
        migrations.AlterUniqueTogether(
            name='floorplan',
            unique_together=set([('location', 'floor')]),
        ),
    ]



from django.test import TestCase

from ..models import FloorPlan, Location, ObjectLocation
from .base.test_models import BaseTestModels
from .testdeviceapp.models import Device


class TestModels(BaseTestModels, TestCase):
    object_model = Device
    location_model = Location
    floorplan_model = FloorPlan
    object_location_model = ObjectLocation



import json

from django.urls import reverse

from .. import TestAdminMixin, TestLociMixin


class BaseTestAdmin(TestAdminMixin, TestLociMixin):
    def test_location_list(self):
        self._login_as_admin()
        self._create_location(name='test-admin-location-1')
        r = self.client.get(reverse('admin:django_loci_location_changelist'))
        self.assertContains(r, 'test-admin-location-1')

    def test_floorplan_list(self):
        self._login_as_admin()
        self._create_floorplan()
        self._create_location()
        r = self.client.get(reverse('admin:django_loci_floorplan_changelist'))
        self.assertContains(r, '1st floor')

    def test_location_json_view(self):
        self._login_as_admin()
        loc = self._create_location()
        r = self.client.get(reverse('admin:django_loci_location_json', args=[loc.pk]))
        expected = {'name': loc.name,
                    'address': loc.address,
                    'type': loc.type,
                    'is_mobile': loc.is_mobile,
                    'geometry': json.loads(loc.geometry.json)}
        self.assertDictEqual(r.json(), expected)

    def test_location_floorplan_json_view(self):
        self._login_as_admin()
        fl = self._create_floorplan()
        r = self.client.get(reverse('admin:django_loci_location_floorplans_json', args=[fl.location.pk]))
        expected = {'choices': [
            {'id': str(fl.pk),
             'str': str(fl),
             'floor': fl.floor,
             'image': fl.image.url,
             'image_width': fl.image.width,
             'image_height': fl.image.height}
        ]}
        self.assertDictEqual(r.json(), expected)



from channels.test import WSClient

from .. import TestAdminMixin, TestLociMixin
from ...channels.base import _get_object_or_none


class BaseTestChannels(TestAdminMixin, TestLociMixin):
    def setUp(self):
        client = WSClient()
        self.client = client

    def test_object_or_none(self):
        result = _get_object_or_none(self.location_model, pk=1)
        self.assertEqual(result, None)
        plausible_pk = self.location_model().pk
        result = _get_object_or_none(self.location_model, pk=plausible_pk)
        self.assertEqual(result, None)

    def _test_ws_add(self, pk=None, user=None):
        if not pk:
            location = self._create_location(is_mobile=True)
            ol = self._create_object_location(location=location)
            pk = ol.location.pk
        path = '/loci/location/{0}/'.format(pk)
        if user:
            self.client.force_login(user)
        self.client.send_and_consume('websocket.connect', path=path)
        return {'ol': ol, 'path': path}

    def test_ws_add_unauthenticated(self):
        try:
            self._test_ws_add()
        except AssertionError as e:
            self.assertIn('Connection rejected', str(e))
        else:
            self.fail('AssertionError not raised')

    def test_connect_and_disconnect(self):
        res = self._test_ws_add(user=self._create_admin())
        self.assertEqual(self.client.receive(), None)
        self.client.send_and_consume('websocket.disconnect', path=res['path'])

    def test_ws_add_not_staff(self):
        user = self.user_model.objects.create_user(username='user',
                                                   password='password',
                                                   email='test@test.org')
        try:
            self._test_ws_add(user=user)
        except AssertionError as e:
            self.assertIn('Connection rejected', str(e))
        else:
            self.fail('AssertionError not raised')

    def test_ws_add_404(self):
        pk = self.location_model().pk
        admin = self._create_admin()
        try:
            self._test_ws_add(pk=pk, user=admin)
        except AssertionError as e:
            self.assertIn('Connection rejected', str(e))
        else:
            self.fail('AssertionError not raised')

    def test_location_update(self):
        res = self._test_ws_add(user=self._create_admin())
        res['ol'].location.geometry = 'POINT (12.513124 41.897903)'
        res['ol'].location.save()
        result = self.client.receive()
        self.assertIsInstance(result, dict)
        self.assertDictEqual(result, {'type': 'Point', 'coordinates': [12.513124, 41.897903]})
